@@ -1,7 +1,7 @@
 package Algorithm::BoostedDecisionTree;
 
 #--------------------------------------------------------------------------------------
-# Copyright (c) 2015 Avinash Kak. All rights reserved.  This program is free
+# Copyright (c) 2016 Avinash Kak. All rights reserved.  This program is free
 # software.  You may modify and/or distribute it under the same terms as Perl itself.
 # This copyright notice must remain attached to the file.
 #
@@ -15,10 +15,10 @@ use lib 'blib/lib', 'blib/arch';
 use strict;
 use warnings;
 use Carp;
-use Algorithm::DecisionTree 3.20;
+use Algorithm::DecisionTree 3.30;
 use List::Util qw(reduce min max);
 
-our $VERSION = '3.20';
+our $VERSION = '3.30';
 
 @BoostedDecisionTree::ISA = ('Algorithm::DecisionTree');
 
@@ -35,7 +35,7 @@ sub new {
     delete $dtargs{how_many_stages};
     my $instance = Algorithm::DecisionTree->new(%dtargs);
     bless $instance, $class;
-    $instance->{_how_many_stages}             =  $args{how_many_stages} || undef;
+    $instance->{_how_many_stages}              =  $args{how_many_stages} || undef;
     $instance->{_stagedebug}                   =  $args{stagedebug} || 0;
     $instance->{_training_samples}             =  {map {$_ => []} 0..$args{how_many_stages}};
     $instance->{_all_trees}                    =  {map {$_ => Algorithm::DecisionTree->new(%dtargs)} 0..$args{how_many_stages}};
@@ -48,61 +48,49 @@ sub new {
     bless $instance, $class;
 }
 
-
 sub get_training_data_for_base_tree {
     my $self = shift;
-    my $numregex =  '[+-]?\ *(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?';
-    my $filename = $self->{_training_datafile} || die "you did not specify a file for the training data";
-    open FILEIN, $filename or die "Unable to open $filename: $!";
-    die("Aborted. get_training_data_csv() is only for CSV files") unless $filename =~ /\.csv$/;
-    my $class_name_in_column = $self->{_csv_class_column_index} - 1;   # subtract 1 because first col has labels
-    my @all_data =  <FILEIN>;
-    my %data_hash = ();
-    my @csv_headers;
-    foreach my $record (@all_data) {
-        next if $record =~ /^#/;     
-        next if $record =~ /^[ ]*\r?\n?$/;
-        next if $record =~ /^[\s=]*$/;
-        $record =~ s/\r?\n?$//;
-        if ($record =~ /^\"\"/) {
-            @csv_headers =  grep $_, map {$_ =~ s/^\"|\"$//g; $_} map {$_ =~ s/^\s*|\s*$//g; $_} split /,/, $record;
-        } else {
-            my @fields =  map {$_ =~ s/^\"|\"$//g; $_} map {$_ =~ s/^\s*|\s*$//g; $_} split /,/, $record;
-            my @fields_after_first = @fields[1..$#fields]; 
-            $data_hash{$fields[0]} = \@fields_after_first;
+    die("Aborted. get_training_data_csv() is only for CSV files") unless $self->{_training_datafile} =~ /\.csv$/;
+    my %class_names = ();
+    my %all_record_ids_with_class_labels;
+    my $firstline;
+    my %data_hash;
+    $|++;
+    open FILEIN, $self->{_training_datafile};
+    my $record_index = 0;
+    my $firsetline;
+    while (<FILEIN>) {
+        next if /^[ ]*\r?\n?$/;
+        $_ =~ s/\r?\n?$//;
+        my $record = cleanup_csv($_);
+        if ($record_index == 0) {
+            $firstline = $record;
+            $record_index++;
+            next;
         }
+        my @parts = split /,/, $record;
+        my $classname = $parts[$self->{_csv_class_column_index}];
+        $class_names{$classname} = 1;
+        my $record_label = shift @parts;
+        $record_label  =~ s/^\s*\"|\"\s*$//g;
+        $data_hash{$record_label} = \@parts;
+        $all_record_ids_with_class_labels{$record_label} = $classname;
+        print "." if $record_index % 10000 == 0;
+        $record_index++;
     }
-    die 'Aborted. The first row of CSV file must begin with "" and then list the feature names and the header label you want to use for the column that will hold class labels for data sample records'
-        unless @csv_headers;
-    my @field_names = @csv_headers;
-    print "\nfield names: @field_names\n" if $self->{_debug2};
-    my $class_column_heading = $field_names[$class_name_in_column];
-    my @feature_names = map {$field_names[$_-1]} @{$self->{_csv_columns_for_features}};
-    print "\nfeature names: @feature_names\n" if $self->{_debug2};
-    $class_column_heading =~ s/^\s*\"|\"\s*$//g;
-    my %class_for_sample_hash = ();
-    my %feature_values_for_samples_hash = ();
-    foreach my $key (keys %data_hash) {
-        next if $key =~ /^\"\"$/;
-        my $cleanedup = $key;
-        $cleanedup =~ s/^\s*\"|\"\s*$//g;
-        my $which_class = $data_hash{$key}[$class_name_in_column];
-        $which_class  =~ s/^\s*\"|\"\s*$//g;
-        $class_for_sample_hash{"sample_$cleanedup"} = "$class_column_heading=$which_class";
-        my @features_and_values_list = ();
-        foreach my $i (@{$self->{_csv_columns_for_features}}) {
-            my $feature_column_header = $field_names[$i-1];
-            my $feature_val = $data_hash{$key}->[$i-1];
-            $feature_val  =~ s/^\s*\"|\"\s*$//g;
-            $feature_val = sprintf("%.1f",$feature_val) if $feature_val =~ /^\d+$/;
-            push @features_and_values_list,  "$feature_column_header=$feature_val";
-        }
-        $feature_values_for_samples_hash{"sample_" . $cleanedup} = \@features_and_values_list;
-    }
-    my @sample_names = keys %feature_values_for_samples_hash;
+    close FILEIN;    
+    $|--;
+    $self->{_how_many_total_training_samples} = $record_index - 1;  # must subtract 1 for the header record
+    print "\n\nTotal number of training samples: $self->{_how_many_total_training_samples}\n" if $self->{_debug1};
+    my @all_feature_names =   split /,/, substr($firstline, index($firstline,','));
+    my $class_column_heading = $all_feature_names[$self->{_csv_class_column_index}];
+    my @feature_names = map {$all_feature_names[$_]} @{$self->{_csv_columns_for_features}};
+    my %class_for_sample_hash = map {"sample_" . $_  =>  "$class_column_heading=" . $data_hash{$_}->[$self->{_csv_class_column_index} - 1 ] } keys %data_hash;
+    my @sample_names = map {"sample_$_"} keys %data_hash;
+    my %feature_values_for_samples_hash = map {my $sampleID = $_; "sample_" . $sampleID  =>  [map {my $fname = $all_feature_names[$_]; $fname . "=" . eval{$data_hash{$sampleID}->[$_-1] =~ /^\d+$/ ? sprintf("%.1f", $data_hash{$sampleID}->[$_-1] ) : $data_hash{$sampleID}->[$_-1] } } @{$self->{_csv_columns_for_features}} ] }  keys %data_hash;    
+    my %features_and_values_hash = map { my $a = $_; {$all_feature_names[$a] => [  map {my $b = $_; $b =~ /^\d+$/ ? sprintf("%.1f",$b) : $b} map {$data_hash{$_}->[$a-1]} keys %data_hash ]} } @{$self->{_csv_columns_for_features}};     
     my @all_class_names =  sort keys %{ {map {$_ => 1} values %class_for_sample_hash } };
     $self->{_number_of_training_samples} = scalar @sample_names;
-    my %features_and_values_hash = map { my $a = $_; {$csv_headers[$a-1] => [  map {my $b = $_; $b =~ /^\d+$/ ? sprintf("%.1f",$b) : $b} map {$data_hash{$_}->[$a-1]} keys %data_hash ]} } @{$self->{_csv_columns_for_features}};     
     if ($self->{_debug2}) {
         print "\nDisplaying features and their values for entire training data:\n\n";
         foreach my $fname (keys  %features_and_values_hash) {         
@@ -112,6 +100,7 @@ sub get_training_data_for_base_tree {
     my %features_and_unique_values_hash = ();
     my %feature_values_how_many_uniques_hash  =  ();
     my %numeric_features_valuerange_hash   =   ();
+    my $numregex =  '[+-]?\ *(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?';
     foreach my $feature (keys %features_and_values_hash) {
         my %seen = ();
         my @unique_values_for_feature =  grep {$_ if $_ ne 'NA' && !$seen{$_}++} @{$features_and_values_hash{$feature}};
@@ -432,7 +421,7 @@ sub get_all_class_names {
 }
 
 
-################################################### Utility Routines #################################################
+################################################ Utility Routines ##############################################
 
 # checks whether an element is in an array:
 sub contained_in {
@@ -491,6 +480,34 @@ sub check_for_illegal_params {
     return $found_match_flag;
 }
 
-1;
+##  Introduced in Version 3.21, I wrote this function in response to a need to
+##  create a decision tree for a very large national econometric database.  The
+##  fields in the CSV file for this database are allowed to be double quoted and such
+##  fields may contain commas inside them.  This function also replaces empty fields
+##  with the generic string 'NA' as a shorthand for "Not Available".  IMPORTANT: This
+##  function skips over the first field in each record.  It is assumed that the first
+##  field in the first record that defines the feature names is the empty string ("")
+##  and the same field in all other records is an ID number for the record.
+sub cleanup_csv {
+    my $line = shift;
+    $line =~ tr/()[]{}/      /;
+    my @double_quoted = substr($line, index($line,',')) =~ /\"[^\"]+\"/g;
+    for (@double_quoted) {
+        my $item = $_;
+        $item = substr($item, 1, -1);
+        $item =~ s/^s+|,|\s+$//g;
+        $item = join '_',  split /\s+/, $item;
+        substr($line, index($line, $_), length($_)) = $item;
+    }
+    my @white_spaced = $line =~ /,\s*[^,]+\s+[^,]+\s*,/g;
+    for (@white_spaced) {
+        my $item = $_;
+        $item = substr($item, 0, -1);
+        $item = join '_',  split /\s+/, $item unless $item =~ /,\s+$/;
+        substr($line, index($line, $_), length($_)) = "$item,";
+    }
+    $line =~ s/,\s*(?=,)/,NA/g;
+    return $line;
+}
 
-    
+1;
