@@ -1,7 +1,7 @@
 package Algorithm::DecisionTreeWithBagging;
 
 #--------------------------------------------------------------------------------------
-# Copyright (c) 2015 Avinash Kak. All rights reserved.  This program is free
+# Copyright (c) 2016 Avinash Kak. All rights reserved.  This program is free
 # software.  You may modify and/or distribute it under the same terms as Perl itself.
 # This copyright notice must remain attached to the file.
 #
@@ -15,9 +15,9 @@ use lib 'blib/lib', 'blib/arch';
 use strict;
 use warnings;
 use Carp;
-use Algorithm::DecisionTree 3.0;
+use Algorithm::DecisionTree 3.30;
 
-our $VERSION = '3.0';
+our $VERSION = '3.30';
 
 ############################################   Constructor  ##############################################
 
@@ -39,7 +39,7 @@ sub new {
         _debug1                       =>  $args{debug1} || 0,
         _number_of_training_samples   =>  undef,
         _segmented_training_data      =>  {},
-        _all_trees                    =>  {map {$_ => Algorithm::DecisionTree->new(%dtargs)} 0..$args{how_many_bags}},
+        _all_trees                    =>  {map {$_ => Algorithm::DecisionTree->new(%dtargs)} 0..$args{how_many_bags} - 1},
         _root_nodes                   =>  [],
         _bag_sizes                    =>  [],
         _classifications              =>  undef,
@@ -48,56 +48,45 @@ sub new {
 
 sub get_training_data_for_bagging {
     my $self = shift;
-    my $numregex =  '[+-]?\ *(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?';
-    my $filename = $self->{_training_datafile} || die "you did not specify a file for the training data";
-    open FILEIN, $filename or die "Unable to open $filename: $!";
-    die("Aborted. get_training_data_csv() is only for CSV files") unless $filename =~ /\.csv$/;
-    my $class_name_in_column = $self->{_csv_class_column_index} - 1;   # subtract 1 because first col has labels
-    my @all_data =  <FILEIN>;
-    my %data_hash = ();
-    my @csv_headers;
-    foreach my $record (@all_data) {
-        next if $record =~ /^#/;     
-        next if $record =~ /^[ ]*\r?\n?$/;
-        next if $record =~ /^[\s=]*$/;
-        $record =~ s/\r?\n?$//;
-        if ($record =~ /^\"\"/) {
-            @csv_headers =  grep $_, map {$_ =~ s/^\"|\"$//g; $_} map {$_ =~ s/^\s*|\s*$//g; $_} split /,/, $record;
-        } else {
-#            my @fields =  map {$_ =~ s/^\"|\"$//; $_} map {$_ =~ s/^\s*|\s*$//; $_} split /,/, $record;
-            my @fields =  map {$_ =~ s/^\"|\"$//g; $_} map {$_ =~ s/^\s*|\s*$//g; $_} split /,/, $record;
-            my @fields_after_first = @fields[1..$#fields]; 
-            $data_hash{$fields[0]} = \@fields_after_first;
+    die("Aborted. get_training_data_csv() is only for CSV files") unless $self->{_training_datafile} =~ /\.csv$/;
+    my %class_names = ();
+    my %all_record_ids_with_class_labels;
+    my $firstline;
+    my %data_hash;
+    $|++;
+    open FILEIN, $self->{_training_datafile} or die "Unable to open $self->{_training_datafile} $!";
+    my $record_index = 0;
+    my $firsetline;
+    while (<FILEIN>) {
+        next if /^[ ]*\r?\n?$/;
+        $_ =~ s/\r?\n?$//;
+        my $record = cleanup_csv($_);
+        if ($record_index == 0) {
+            $firstline = $record;
+            $record_index++;
+            next;
         }
+        my @parts = split /,/, $record;
+        my $classname = $parts[$self->{_csv_class_column_index}];
+        $class_names{$classname} = 1;
+        my $record_label = shift @parts;
+        $record_label  =~ s/^\s*\"|\"\s*$//g;
+        $data_hash{$record_label} = \@parts;
+        $all_record_ids_with_class_labels{$record_label} = $classname;
+        print "." if $record_index % 10000 == 0;
+        $record_index++;
     }
-    die 'Aborted. The first row of CSV file must begin with "" and then list the feature names and the header label you want to use for the column that will hold class labels for data sample records'
-        unless @csv_headers;
-    my @field_names = @csv_headers;
-    print "\nfield names: @field_names\n" if $self->{_debug2};
-    my $class_column_heading = $field_names[$class_name_in_column];
-    my @feature_names = map {$field_names[$_-1]} @{$self->{_csv_columns_for_features}};
-    print "\nfeature names: @feature_names\n" if $self->{_debug2};
-    $class_column_heading =~ s/^\s*\"|\"\s*$//g;
-    my %class_for_sample_hash = ();
-    my %feature_values_for_samples_hash = ();
-    foreach my $key (keys %data_hash) {
-        next if $key =~ /^\"\"$/;
-        my $cleanedup = $key;
-        $cleanedup =~ s/^\s*\"|\"\s*$//g;
-        my $which_class = $data_hash{$key}[$class_name_in_column];
-        $which_class  =~ s/^\s*\"|\"\s*$//g;
-        $class_for_sample_hash{"sample_$cleanedup"} = "$class_column_heading=$which_class";
-        my @features_and_values_list = ();
-        foreach my $i (@{$self->{_csv_columns_for_features}}) {
-            my $feature_column_header = $field_names[$i-1];
-            my $feature_val = $data_hash{$key}->[$i-1];
-            $feature_val  =~ s/^\s*\"|\"\s*$//g;
-            $feature_val = sprintf("%.1f",$feature_val) if $feature_val =~ /^\d+$/;
-            push @features_and_values_list,  "$feature_column_header=$feature_val";
-        }
-        $feature_values_for_samples_hash{"sample_" . $cleanedup} = \@features_and_values_list;
-    }
-    my @sample_names = keys %feature_values_for_samples_hash;
+    close FILEIN;    
+    $|--;
+    $self->{_how_many_total_training_samples} = $record_index - 1;  # must subtract 1 for the header record
+    print "\n\nTotal number of training samples: $self->{_how_many_total_training_samples}\n" if $self->{_debug1};
+    my @all_feature_names =   split /,/, substr($firstline, index($firstline,','));
+    my $class_column_heading = $all_feature_names[$self->{_csv_class_column_index}];
+#    my @all_class_names = sort map {"$class_column_heading=$_"} keys %class_names;
+    my @feature_names = map {$all_feature_names[$_]} @{$self->{_csv_columns_for_features}};
+    my %class_for_sample_hash = map {"sample_" . $_  =>  "$class_column_heading=" . $data_hash{$_}->[$self->{_csv_class_column_index} - 1 ] } keys %data_hash;
+    my @sample_names = map {"sample_$_"} keys %data_hash;
+    my %feature_values_for_samples_hash = map {my $sampleID = $_; "sample_" . $sampleID  =>  [map {my $fname = $all_feature_names[$_]; $fname . "=" . eval{$data_hash{$sampleID}->[$_-1] =~ /^\d+$/ ? sprintf("%.1f", $data_hash{$sampleID}->[$_-1] ) : $data_hash{$sampleID}->[$_-1] } } @{$self->{_csv_columns_for_features}} ] }  keys %data_hash;    
     $self->{_number_of_training_samples} = scalar @sample_names;
     fisher_yates_shuffle(\@sample_names);
     print "\nsample names for all samples: @sample_names\n" if $self->{_debug2};
@@ -148,14 +137,14 @@ sub get_training_data_for_bagging {
             }
         }
     }
-    my %features_and_values_hash = map { my $a = $_; {$csv_headers[$a-1] => [  map {my $b = $_; $b =~ /^\d+$/ ? sprintf("%.1f",$b) : $b} map {$data_hash{$_}->[$a-1]} keys %data_hash ]} } @{$self->{_csv_columns_for_features}};     
+    my %features_and_values_hash = map { my $a = $_; {$all_feature_names[$a] => [  map {my $b = $_; $b =~ /^\d+$/ ? sprintf("%.1f",$b) : $b} map {$data_hash{$_}->[$a-1]} keys %data_hash ]} } @{$self->{_csv_columns_for_features}};     
     if ($self->{_debug2}) {
         print "\nDisplaying features and their values for entire training data:\n\n";
         foreach my $fname (keys  %features_and_values_hash) {         
             print "        $fname    =>  @{$features_and_values_hash{$fname}}\n";
         }
     }
-    my %features_and_values_hash_bags =  map { my $c = $_; { $c =>  { map { my $d = $_; {$csv_headers[$d-1] => [ sort {$a cmp $b} map {my $f = $_; $f =~ /^\d+$/ ? sprintf("%.1f",$f) : $f} map {$data_hash{sample_index($_)}->[$d-1]} @{$data_sample_bags[$c]} ] } } @{$self->{_csv_columns_for_features}} } } } 0 .. $self->{_how_many_bags} - 1;         
+    my %features_and_values_hash_bags =  map { my $c = $_; { $c =>  { map { my $d = $_; {$all_feature_names[$d] => [ sort {$a cmp $b} map {my $f = $_; $f =~ /^\d+$/ ? sprintf("%.1f",$f) : $f} map {$data_hash{sample_index($_)}->[$d-1]} @{$data_sample_bags[$c]} ] } } @{$self->{_csv_columns_for_features}} } } } 0 .. $self->{_how_many_bags} - 1;         
     if ($self->{_debug2}) {
         print "\nDisplaying features and their values in each bag:\n\n";
         foreach my $bag_index (keys  %features_and_values_hash_bags) {           
@@ -171,6 +160,7 @@ sub get_training_data_for_bagging {
     my %numeric_features_valuerange_hash_bags   =   map {$_ => {}} 0 .. $self->{_how_many_bags} - 1;
     my %feature_values_how_many_uniques_hash_bags   =   map {$_ => {}} 0 .. $self->{_how_many_bags} - 1;
     my %features_and_unique_values_hash_bags   =   map {$_ => {}} 0 .. $self->{_how_many_bags} - 1;
+    my $numregex =  '[+-]?\ *(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?';
     foreach my $i (0 .. $self->{_how_many_bags} - 1) {
         foreach my $feature (keys %{$features_and_values_hash_bags{$i}}) {
             my %seen = ();
@@ -187,7 +177,7 @@ sub get_training_data_for_bagging {
     }
     if ($self->{_debug2}) {
         print "\nDisplaying value ranges for numeric features in each bag:\n\n";
-        foreach my $bag_index (keys  %numeric_features_valuerange_hash_bags) {                  #KEEP THIS
+        foreach my $bag_index (keys  %numeric_features_valuerange_hash_bags) {        
             my %keyval = %{$numeric_features_valuerange_hash_bags{$bag_index}};
             print "\nFor bag $bag_index  =>:\n";
             foreach my $fname (keys %keyval) {
@@ -195,7 +185,7 @@ sub get_training_data_for_bagging {
             }
         }
         print "\nDisplaying number of unique values for each features in each bag:\n\n";
-        foreach my $bag_index (keys  %feature_values_how_many_uniques_hash_bags) {                  #KEEP THIS
+        foreach my $bag_index (keys  %feature_values_how_many_uniques_hash_bags) {    
             my %keyval = %{$feature_values_how_many_uniques_hash_bags{$bag_index}};
             print "\nFor bag $bag_index  =>:\n";
             foreach my $fname (keys %keyval) {
@@ -408,4 +398,34 @@ sub fisher_yates_shuffle {
     }
 }
 
+##  Introduced in Version 3.21, I wrote this function in response to a need to
+##  create a decision tree for a very large national econometric database.  The
+##  fields in the CSV file for this database are allowed to be double quoted and such
+##  fields may contain commas inside them.  This function also replaces empty fields
+##  with the generic string 'NA' as a shorthand for "Not Available".  IMPORTANT: This
+##  function skips over the first field in each record.  It is assumed that the first
+##  field in the first record that defines the feature names is the empty string ("")
+##  and the same field in all other records is an ID number for the record.
+sub cleanup_csv {
+    my $line = shift;
+    $line =~ tr/()[]{}/      /;
+    my @double_quoted = substr($line, index($line,',')) =~ /\"[^\"]+\"/g;
+    for (@double_quoted) {
+        my $item = $_;
+        $item = substr($item, 1, -1);
+        $item =~ s/^s+|,|\s+$//g;
+        $item = join '_',  split /\s+/, $item;
+        substr($line, index($line, $_), length($_)) = $item;
+    }
+    my @white_spaced = $line =~ /,\s*[^,]+\s+[^,]+\s*,/g;
+    for (@white_spaced) {
+        my $item = $_;
+        $item = substr($item, 0, -1);
+        $item = join '_',  split /\s+/, $item unless $item =~ /,\s+$/;
+        substr($line, index($line, $_), length($_)) = "$item,";
+    }
+    $line =~ s/,\s*(?=,)/,NA/g;
+    return $line;
+}
+
 1;
